@@ -5,6 +5,7 @@
  */
 
 # include "jam.h"
+# include "debug.h"
 
 # include "lists.h"
 # include "parse.h"
@@ -21,6 +22,8 @@
 # include "make.h"
 # include "hdrmacro.h"
 # include "compile.h"
+# include "native.h"
+# include "variable.h"
 # include <ctype.h>
 
 /*
@@ -49,9 +52,13 @@
 # define P0 (PARSE *)0
 # define C0 (char *)0
 
+# if defined( OS_NT ) || defined( OS_CYGWIN )
+LIST* builtin_system_registry( PARSE *parse, FRAME *frame );
+# endif
+
 int glob( char *s, char *c );
 
-static void lol_build( LOL* lol, char** elements );
+void lol_build( LOL* lol, char** elements );
 void backtrace( FRAME *frame );
 void backtrace_line( FRAME *frame );
 void print_source_line( PARSE* p );
@@ -91,10 +98,14 @@ load_builtins()
       bind_builtin( "ECHO" ,
                     builtin_echo, 0, 0 ) ) );
 
-    duplicate_rule( "exit" ,
-    duplicate_rule( "Exit" ,
-      bind_builtin( "EXIT" ,
-                    builtin_exit, 0, 0 ) ) );
+
+    {
+        char * args[] = { "message", "*", ":", "result-value", "?", 0 };
+        duplicate_rule( "exit" ,
+        duplicate_rule( "Exit" ,
+          bind_builtin( "EXIT" ,
+                        builtin_exit, 0, args ) ) );
+    }
 
     {
         char * args[] = { "directories", "*", ":", "patterns", "*", ":", "case-insensitive", "?", 0 };
@@ -104,10 +115,22 @@ load_builtins()
             );
     }
 
+    {
+        char * args[] = { "patterns", "*", 0 };
+        bind_builtin( "GLOB-RECURSIVELY" , builtin_glob_recursive, 0, args );
+    }
+
+
     duplicate_rule( "Includes" ,
       bind_builtin( "INCLUDES" ,
                     builtin_depends, 1, 0 ) );
 
+    {
+        char * args[] = { "targets", "*", ":", "targets-to-rebuild", "*", 0 };
+        bind_builtin( "REBUILDS" ,
+                      builtin_rebuilds, 0, args );
+    }
+    
     duplicate_rule( "Leaves" ,
       bind_builtin( "LEAVES" ,
                     builtin_flags, T_FLAG_LEAVES, 0 ) );
@@ -133,6 +156,13 @@ load_builtins()
       bind_builtin( "TEMPORARY" ,
                     builtin_flags, T_FLAG_TEMP, 0 ) );
 
+    {
+        char * args[] = { "targets", "*", 0 };
+        bind_builtin(
+            "ISFILE",
+            builtin_flags, T_FLAG_ISFILE, 0 );
+    }
+
     duplicate_rule( "HdrMacro" ,
       bind_builtin( "HDRMACRO" ,
                     builtin_hdrmacro, 0, 0 ) );
@@ -241,10 +271,134 @@ load_builtins()
       }
 
       {
-          char * args[] = { "path", 0 };
+          char * args[] = { "path_parts", "*", 0 };
           bind_builtin( "NORMALIZE_PATH",
-                        builtin_normalize_path, 0, args );
+              builtin_normalize_path, 0, args );
       }
+
+      {
+          char * args[] = { "args", "*", 0 };
+          bind_builtin( "CALC",
+              builtin_calc, 0, args );
+      }
+
+      {
+          char * args[] = { "module", ":", "rule", 0 };
+          bind_builtin( "NATIVE_RULE",
+              builtin_native_rule, 0, args );
+      }
+
+      {
+          char * args[] = { "module", ":", "rule", ":", "version", 0 };
+          bind_builtin( "HAS_NATIVE_RULE",
+              builtin_has_native_rule, 0, args );
+      }
+
+
+      {
+          char * args[] = { "module", "*", 0 };
+          bind_builtin( "USER_MODULE",
+              builtin_user_module, 0, args );
+      }
+
+      {
+          char * args[] = { 0 };
+          bind_builtin( "NEAREST_USER_LOCATION",
+              builtin_nearest_user_location, 0, args );
+      }
+
+      {
+          char * args[] = { "file", 0 };
+          bind_builtin( "CHECK_IF_FILE",
+                        builtin_check_if_file, 0, args );
+      }
+
+#ifdef HAVE_PYTHON
+      {
+          char * args[] = { "python-module", ":", "function", ":", 
+                            "jam-module", ":", "rule-name", 0 };
+          bind_builtin( "PYTHON_IMPORT_RULE",
+              builtin_python_import_rule, 0, args );
+      }
+#endif
+
+# if defined( OS_NT ) || defined( OS_CYGWIN )
+      {
+          char * args[] = { "key_path", ":", "data", "?", 0 };
+          bind_builtin( "W32_GETREG",
+              builtin_system_registry, 0, args );
+      }
+# endif
+
+      {
+          char * args[] = { "command", ":", "*", 0 };
+          bind_builtin( "SHELL",
+              builtin_shell, 0, args );
+          bind_builtin( "COMMAND",
+              builtin_shell, 0, args );
+      }
+
+      /* Initialize builtin modules */
+      init_set();
+      init_path();
+      init_regex();
+      init_property_set();
+      init_sequence();
+      init_order();
+}
+
+/*
+* builtin_calc() - CALC rule
+*
+* The CALC rule performs simple mathematical operations on two arguments.
+*/
+
+LIST *
+builtin_calc(
+    PARSE *parse,
+    FRAME *frame )
+{
+    LIST *arg = lol_get( frame->args, 0 );
+
+    LIST *result = 0;
+    long lhs_value;
+    long rhs_value;
+    long result_value;
+    char buffer [16];
+    const char* lhs;
+    const char* op;
+    const char* rhs;
+
+    if (arg == 0) return L0;
+    lhs = arg->string;
+
+    arg = list_next( arg );
+    if (arg == 0) return L0;
+    op = arg->string;
+
+    arg = list_next( arg );
+    if (arg == 0) return L0;
+    rhs = arg->string;
+
+    lhs_value = atoi (lhs);
+    rhs_value = atoi (rhs);
+
+    if (strcmp ("+", op) == 0)
+    {
+        result_value = lhs_value + rhs_value;
+    }
+    else if (strcmp ("-", op) == 0)
+    {
+        result_value = lhs_value - rhs_value;
+    }
+    else
+    {
+        return L0;
+    }
+
+    sprintf (buffer, "%ld", result_value);
+    result = list_new( result, newstr( buffer ) );
+    return result;
 }
 
 /*
@@ -262,7 +416,6 @@ builtin_depends(
 {
 	LIST *targets = lol_get( frame->args, 0 );
 	LIST *sources = lol_get( frame->args, 1 );
-	int which = parse->num;
 	LIST *l;
 
 	for( l = targets; l; l = list_next( l ) )
@@ -285,6 +438,39 @@ builtin_depends(
 	    t->depends = targetlist( t->depends, sources );
 	}
 
+    /* Enter reverse links */
+	for( l = sources; l; l = list_next( l ) )
+	{
+	    TARGET *s = bindtarget( l->string );
+        s->dependents = targetlist( s->dependents, targets );
+    }
+
+	return L0;
+}
+
+/*
+ * builtin_rebuilds() - REBUILDS rule
+ *
+ * The REBUILDS builtin rule appends each of the listed
+ * rebuild-targets in its 2nd argument on the rebuilds list of each of
+ * the listed targets in its first argument.
+ */
+
+LIST *
+builtin_rebuilds(
+	PARSE	*parse,
+	FRAME *frame )
+{
+	LIST *targets = lol_get( frame->args, 0 );
+	LIST *rebuilds = lol_get( frame->args, 1 );
+	LIST *l;
+
+	for( l = targets; l; l = list_next( l ) )
+	{
+	    TARGET *t = bindtarget( l->string );
+	    t->rebuilds = targetlist( t->rebuilds, rebuilds );
+	}
+
 	return L0;
 }
 
@@ -314,13 +500,20 @@ builtin_echo(
 
 LIST *
 builtin_exit(
-	PARSE	*parse,
-	FRAME *frame )
+    PARSE   *parse,
+    FRAME   *frame )
 {
-	list_print( lol_get( frame->args, 0 ) );
-	printf( "\n" );
-	exit( EXITBAD ); /* yeech */
-	return L0;
+    list_print( lol_get( frame->args, 0 ) );
+    printf( "\n" );
+    if ( lol_get( frame->args, 1 ) )
+    {
+        exit ( atoi( lol_get( frame->args, 1 )->string ) );
+    }
+    else
+    {
+        exit( EXITBAD ); /* yeech */
+    }
+    return L0;
 }
 
 /*
@@ -368,21 +561,37 @@ builtin_glob_back(
     int status,
     time_t  time )
 {
+    PROFILE_ENTER(BUILTIN_GLOB_BACK);
+    
     struct globbing *globbing = (struct globbing *)closure;
     LIST        *l;
     PATHNAME    f;
     string          buf[1];
-
+    
     /* Null out directory for matching. */
     /* We wish we had file_dirscan() pass up a PATHNAME. */
 
     path_parse( file, &f );
     f.f_dir.len = 0;
+
+    /* For globbing, we unconditionally ignore current and parent
+       directory items. Since they items always exist, there's not
+       reason why caller of GLOB would want to see them.
+       We could also change file_dirscan, but then paths with embedded
+       "." and ".." won't work anywhere.
+    */
+    if (strcmp(f.f_base.ptr, ".") == 0 || strcmp(f.f_base.ptr, "..") == 0)
+    {
+        PROFILE_EXIT(BUILTIN_GLOB_BACK);
+        return;
+    }
+
     string_new( buf );
     path_build( &f, buf, 0 );
 
-    if (globbing->case_insensitive)
+    if (globbing->case_insensitive) {
         downcase_inplace( buf->value );
+    }
 
     for( l = globbing->patterns; l; l = l->next )
     {
@@ -394,6 +603,8 @@ builtin_glob_back(
     }
     
     string_free( buf );
+    
+    PROFILE_EXIT(BUILTIN_GLOB_BACK);
 }
 
 static LIST* downcase_list( LIST *in )
@@ -450,6 +661,160 @@ builtin_glob(
     return globbing.results;
 }
 
+static int has_wildcards(const char* str)
+{
+    size_t index = strcspn(str, "[]*?");
+    if (str[index] == '\0')
+        return 0;
+    else
+        return 1;
+}
+
+/** If 'file' exists, append 'file' to 'list'.
+    Returns 'list'.
+*/
+static LIST* append_if_exists(LIST* list, char* file)
+{
+    time_t time;
+    timestamp(file, &time);
+    if (time > 0)
+        return list_new(list, newstr(file));
+    else
+        return list;        
+}
+
+LIST* glob1(char* dirname, char* pattern)
+{
+    LIST* plist = list_new(L0, pattern);
+    struct globbing globbing;
+
+    globbing.results = L0;
+    globbing.patterns = plist;
+    
+    globbing.case_insensitive
+# if defined( OS_NT ) || defined( OS_CYGWIN )
+       = plist;  /* always case-insensitive if any files can be found */
+# else 
+       = L0;
+# endif
+
+    if ( globbing.case_insensitive )
+    {
+        globbing.patterns = downcase_list( plist );
+    }
+    
+    file_dirscan( dirname, builtin_glob_back, &globbing );
+
+    if ( globbing.case_insensitive )
+    {
+        list_free( globbing.patterns );
+    }
+
+    list_free(plist);
+
+    return globbing.results;
+}
+
+
+LIST* glob_recursive(char* pattern)
+{
+    LIST* result = L0;
+
+    /* Check if there's metacharacters in pattern */
+    if (!has_wildcards(pattern))
+    {
+        /* No metacharacters. Check if the path exists. */
+        result = append_if_exists(result, pattern);
+    }        
+    else
+    {
+        /* Have metacharacters in the pattern. Split into dir/name */
+        PATHNAME path[1];
+        path_parse(pattern, path);            
+        
+        if (path->f_dir.ptr)
+        {
+            LIST* dirs = L0;
+            string dirname[1];
+            string basename[1];
+            string_new(dirname);
+            string_new(basename);
+
+            string_append_range(dirname, path->f_dir.ptr, 
+                                path->f_dir.ptr + path->f_dir.len);
+
+            path->f_grist.ptr = 0;
+            path->f_grist.len = 0;
+            path->f_dir.ptr = 0;
+            path->f_dir.len = 0;
+            path_build(path, basename, 0);
+
+            if (has_wildcards(dirname->value))
+            {
+                dirs = glob_recursive(dirname->value);
+            }
+            else
+            {
+                dirs = list_new(dirs, dirname->value);
+            }
+            
+            if (has_wildcards(basename->value))
+            {
+                for(; dirs; dirs = dirs->next)
+                {
+                    result = list_append(result, 
+                                         glob1(dirs->string, basename->value));
+                }
+            }
+            else
+            {
+                string file_string[1];
+                string_new(file_string);
+
+                /** No wildcard in basename. */
+                for(; dirs; dirs = dirs->next)
+                {                                      
+                    path->f_dir.ptr = dirs->string;
+                    path->f_dir.len = strlen(dirs->string);                    
+                    path_build(path, file_string, 0);
+
+                    result = append_if_exists(result, file_string->value);
+
+                    string_truncate(file_string, 0);
+                }
+
+                string_free(file_string);
+            }
+
+            string_free(dirname);
+            string_free(basename);
+        }
+        else
+        {
+            /** No directory, just a pattern. */
+            result = list_append(result, glob1(".", pattern));
+        }
+    }
+
+    return result;
+}
+
+LIST *
+builtin_glob_recursive(
+    PARSE   *parse,
+    FRAME *frame )
+{
+    LIST* result = L0;
+    LIST* l = lol_get( frame->args, 0 );
+
+    for(; l; l = l->next)
+    {
+        result = list_append(result, glob_recursive(l->string));
+    }
+
+    return result;
+}
+
 /*
  * builtin_match() - MATCH rule, regexp matching
  */
@@ -919,7 +1284,7 @@ builtin_sort( PARSE *parse, FRAME *frame )
 
 LIST *builtin_normalize_path( PARSE *parse, FRAME *frame )
 {
-    LIST* arg1 = lol_get( frame->args, 0 );
+    LIST* arg = lol_get( frame->args, 0 );
 
     /* First, we iterate over all '/'-separated elements, starting from
        the end of string. If we see '..', we remove previous path elements.
@@ -932,14 +1297,20 @@ LIST *builtin_normalize_path( PARSE *parse, FRAME *frame )
     char* end;      /* Last character of the part of string still to be processed. */
     char* current;  /* Working pointer. */  
     int dotdots = 0; /* Number of '..' elements seen and not processed yet. */
-    int rooted = arg1->string[0] == '/';
+    int rooted = arg->string[0] == '/';
     char* result;
 
     /* Make a copy of input: we should not change it. */
     string_new(in);
     if (!rooted)
         string_push_back(in, '/');
-    string_append(in, arg1->string);
+    while (arg)
+    {
+        string_append(in, arg->string);
+        arg = list_next(arg);
+        if (arg)
+            string_append(in, "/");
+    }
     
 
     end = in->value + in->size - 1;
@@ -1007,9 +1378,182 @@ LIST *builtin_normalize_path( PARSE *parse, FRAME *frame )
 
 }
 
+LIST *builtin_native_rule( PARSE *parse, FRAME *frame )
+{
+    LIST* module_name = lol_get( frame->args, 0 );    
+    LIST* rule_name = lol_get( frame->args, 1 );    
+
+    module_t* module = bindmodule(module_name->string);
+
+    native_rule_t n, *np = &n;
+    n.name = rule_name->string;
+    if (module->native_rules && hashcheck(module->native_rules, (HASHDATA**)&np))
+    {
+        new_rule_body(module, np->name, np->arguments, np->procedure, 1);
+    }
+    else
+    {
+        backtrace_line( frame->prev );
+        printf( "error: no native rule \"%s\" defined in module \"%s\"\n", 
+                n.name, module->name);
+        backtrace( frame->prev );
+        exit(1);
+    }
+    return L0;    
+}
+
+LIST *builtin_has_native_rule( PARSE *parse, FRAME *frame )
+{
+    LIST* module_name = lol_get( frame->args, 0 );    
+    LIST* rule_name = lol_get( frame->args, 1 );    
+    LIST* version = lol_get( frame->args, 2 );    
+
+    module_t* module = bindmodule(module_name->string);
+
+    native_rule_t n, *np = &n;
+    n.name = rule_name->string;
+    if (module->native_rules && hashcheck(module->native_rules, (HASHDATA**)&np))
+    {
+        int expected_version = atoi(version->string);
+        if (np->version == expected_version)
+            return list_new(0, newstr("true"));
+    }
+    return L0;    
+}
 
 
-static void lol_build( LOL* lol, char** elements )
+LIST *builtin_user_module( PARSE *parse, FRAME *frame )
+{
+    LIST* module_name = lol_get( frame->args, 0 );    
+    for(; module_name; module_name = module_name->next) 
+    {
+        module_t* m = bindmodule( module_name->string);
+        m->user_module = 1;
+    }
+    return L0;
+}
+
+LIST *builtin_nearest_user_location( PARSE *parse, FRAME *frame )
+{
+    LIST* result = 0;
+    FRAME* nearest_user_frame = 
+        frame->module->user_module ? frame : frame->prev_user;
+
+    if (nearest_user_frame)
+    {
+        char* file;
+        int line;
+        char buf[32];
+        get_source_line( nearest_user_frame->procedure, &file, &line );
+        sprintf( buf, "%d", line );
+        result = list_new( result, newstr( file ) );
+        result = list_new( result, newstr( buf ) );
+        return result;
+    }
+    else
+    {
+        return L0;
+    }
+}
+
+LIST *builtin_check_if_file( PARSE *parse, FRAME *frame )
+{
+    LIST* name = lol_get( frame->args, 0 );
+    if (file_is_file(name->string) == 1) {
+        return list_new(0, newstr("true"));
+    } else {
+        return L0;
+    }
+}
+
+
+#ifdef HAVE_PYTHON
+
+LIST *builtin_python_import_rule( PARSE *parse, FRAME *frame )
+{
+    static int first_time = 1;
+   char* python_module = lol_get( frame->args, 0 )->string;        
+   char* python_function = lol_get( frame->args, 1 )->string;        
+   char* jam_module = lol_get( frame->args, 2 )->string;        
+   char* jam_rule = lol_get( frame->args, 3 )->string;        
+
+   PyObject *pName, *pModule, *pDict, *pFunc;
+
+   if (first_time)
+   {
+       /* At the first invocation, we add the value of the
+          global EXTRA_PYTHONPATH to the sys.path Python
+          variable.
+       */
+       LIST* extra = 0;
+       module_t* outer_module = frame->module;
+
+       first_time = 0;
+
+       if ( outer_module != root_module())
+       {
+           exit_module( outer_module );
+           enter_module( root_module());
+       }
+    
+       extra = var_get("EXTRA_PYTHONPATH");
+    
+       if ( outer_module != root_module())
+       {
+            exit_module( root_module());
+            enter_module( outer_module );
+       }
+
+       for(; extra; extra = extra->next)
+       {
+           string buf[1];
+           string_new(buf);
+           string_append(buf, "import sys\nsys.path.append(\"");
+           string_append(buf, extra->string);
+           string_append(buf, "\")\n");
+           PyRun_SimpleString(buf->value);   
+           string_free(buf);               
+       }       
+   }
+
+
+   pName = PyString_FromString(python_module);
+   
+   pModule = PyImport_Import(pName);
+   Py_DECREF(pName);
+
+   if (pModule != NULL) {
+        pDict = PyModule_GetDict(pModule);
+        pFunc = PyDict_GetItemString(pDict, python_function);
+
+        if (pFunc && PyCallable_Check(pFunc)) {
+
+            module_t* m = bindmodule(jam_module);
+            RULE* r = bindrule( jam_rule, m );
+
+            /* Make pFunc owned */
+            Py_INCREF(pFunc);
+
+            r->python_function = pFunc;
+        }
+        else {
+            if (PyErr_Occurred())
+                PyErr_Print();
+            fprintf(stderr, "Cannot find function \"%s\"\n", python_function);
+        }
+        Py_DECREF(pModule);
+    }
+    else {
+        PyErr_Print();
+        fprintf(stderr, "Failed to load \"%s\"\n", python_module);
+    }
+   return L0;
+
+}
+
+#endif
+
+void lol_build( LOL* lol, char** elements )
 {
     LIST* l = L0;
     lol_init( lol );
@@ -1032,3 +1576,170 @@ static void lol_build( LOL* lol, char** elements )
         lol_add( lol, l );
 }
 
+#ifdef HAVE_PYTHON
+
+/** Calls the bjam rule specified by name passed in 'args'.
+    The name is looked up in context of bjam's 'python_interface'
+    module. Returns the list of string retured by the rule.
+*/
+PyObject*
+bjam_call(PyObject* self, PyObject* args)
+{
+    FRAME       inner[1];
+    LIST    *result;
+    PARSE   *p;
+    char*  rulename;
+    
+    /* Build up the list of arg lists */
+
+    frame_init( inner );
+    inner->prev = 0;
+    inner->prev_user = 0;
+    inner->module = bindmodule("python_interface");
+    inner->procedure = 0;
+
+    /* Extract the rule name and arguments from 'args' */
+
+    /* PyTuple_GetItem returns borrowed reference */
+    rulename = PyString_AsString(PyTuple_GetItem(args, 0));
+    {
+        int i = 1;
+        int size = PyTuple_Size(args);
+        for( ; i < size; ++i) {
+            PyObject* a = PyTuple_GetItem(args, i);
+            if (PyString_Check(a))
+            {
+                lol_add(inner->args, 
+                        list_new(0, newstr(PyString_AsString(a))));
+            }
+            else if (PySequence_Check(a))
+            {
+                LIST* l = 0;
+                int s = PySequence_Size(a);
+                int i = 0;
+                for(; i < s; ++i)
+                {
+                    /* PySequence_GetItem returns new reference. */
+                    PyObject* e = PySequence_GetItem(a, i);
+                    l = list_new(l, newstr(PyString_AsString(e)));
+                    Py_DECREF(e);
+                }
+                lol_add(inner->args, l);
+            }                
+        }
+    }
+
+    result = evaluate_rule( rulename, inner );
+
+    frame_free( inner );
+}
+
+/** Accepts three arguments: module name, rule name and Python callable.
+
+    Creates bjam rule with the specified name in the specified module,
+    which will invoke the Python callable.
+*/
+PyObject*
+bjam_import_rule(PyObject* self, PyObject* args)
+{
+    char* module;
+    char* rule;
+    PyObject* func;
+    module_t* m;
+    RULE* r;
+
+    if (!PyArg_ParseTuple(args, "ssO:import_rule", &module, &rule, &func))
+        return NULL;
+    
+    if (!PyCallable_Check(func))
+        return NULL;
+    
+    m = bindmodule(module);
+    r = bindrule(rule, m);
+
+    /* Make pFunc owned */
+    Py_INCREF(func);
+
+    r->python_function = func;
+    return Py_None;
+}
+
+#endif
+
+#ifdef HAVE_POPEN
+#if defined(_MSC_VER) || defined(__BORLANDC__)
+    #define popen _popen
+    #define pclose _pclose
+#endif
+
+LIST *builtin_shell( PARSE *parse, FRAME *frame )
+{
+    LIST* command = lol_get( frame->args, 0 );
+    LIST* result = 0;
+    string s;
+    int ret;
+    char buffer[1024];
+    FILE *p = NULL;
+    int exit_status = -1;
+    int exit_status_opt = 0;
+    int no_output_opt = 0;
+    
+    /* Process the variable args options. */
+    {
+        int a = 1;
+        LIST * arg = lol_get( frame->args, a );
+        while ( arg )
+        {
+            if ( strcmp("exit-status", arg->string) == 0 )
+            {
+                exit_status_opt = 1;
+            }
+            else if ( strcmp("no-output", arg->string) == 0 )
+            {
+                no_output_opt = 1;
+            }
+            arg = lol_get( frame->args, ++a );
+        }
+    }
+
+    string_new( &s );
+
+    fflush(NULL);
+
+    p = popen(command->string, "r");
+    if ( p == NULL )
+        return L0;
+
+    while ( (ret = fread(buffer, sizeof(char), sizeof(buffer)-1, p)) > 0 )
+    {
+        buffer[ret] = 0;
+        if ( ! no_output_opt )
+        {
+            string_append( &s, buffer );
+        }
+    }
+
+    exit_status = pclose(p);
+
+    /* The command output is returned first. */
+    result = list_new( L0, newstr(s.value) );
+    string_free(&s);
+    
+    /* The command exit result next. */
+    if ( exit_status_opt )
+    {
+        sprintf (buffer, "%d", exit_status);
+        result = list_new( result, newstr( buffer ) );
+    }
+    
+    return result;
+}
+
+#else
+
+LIST *builtin_shell( PARSE *parse, FRAME *frame )
+{
+    return L0;
+}
+
+#endif
